	TITLE	'SIO AND 2SIO MODULE FOR CP/M 3.0'

	; CHARIO3.ASM  v.1.0  02/22/2023

	; Written by Wayne Parham  (wayne@parhamdata.com)

	; Character device driver module for Altair computers.
	; Supports SIO and 2SIO async boards and compatibles
	; such as Martin Eberhard's 88-2SIOJP.

	; Note that the CIST function calls ckFlush from the
	; diskette module.  This allows delayed track flushing.

TRUE		EQU	-1
FALSE		EQU	NOT TRUE

	; Equates used by FDC for track flushing while system polling for user input
UNDEF		EQU	0FFH
MINIDSK		EQU	FALSE		;Set TRUE for Altair Mini-Disk (see FDC3.asm)
cHDUNLD		EQU	08H		;head unload
DRVCMD		EQU	09H		;drive command register (out)

BANKED		EQU	FALSE	;<--------------- NON-BANKED VERSION

	; Public labels:
	PUBLIC	?CINIT,?CI,?CO,?CIST,?COST
	PUBLIC	@CTBL

	; External labels:
   IF	BANKED
	EXTRN	@CBNK
	EXTRN	?BNKSL
   ENDIF
	EXTRN	?PMSG
	; FDC external labels:
	EXTRN	bfDrive		;Buffered drive
	EXTRN	ckFlush		;Disk buffer flush routine

	; INCLUDE Z-80 MACROS:
	MACLIB	Z80

	; 2SIO Serial Board Equates 

SIO1CTL	EQU	10h		;1st port on 2SIO board - control register
SIO1DAT	EQU	11h		;1st port on 2SIO board - data register
SIO2CTL	EQU	12h		;2nd port on 2SIO board - control register
SIO2DAT	EQU	13h		;2nd port on 2SIO board - data register
SIORDRF	EQU	01h		;read data register full flag
SIOTDRE	EQU	02h		;transmit data register empty flag

	; SIO Serial Board Equates

SIOCTL	EQU	00h		;SIO board at 0 - control register
SIODAT	EQU	01h		;SIO board at 0 - data register
ACRCTL	EQU	06h		;SIO board at 6 - control register (cassette)
ACRDAT	EQU	07h		;SIO board at 6 - data register (cassette)
SIORCV	EQU	01h		;data received bit (inverted)
SIOXMT	EQU	80h		;read to transmit (inverted)

	; Mode byte equates

MB$INPUT	EQU 0000$0001B	;input device
MB$OUTPUT	EQU 0000$0010B	;output device
MB$IN$OUT	EQU MB$INPUT+MB$OUTPUT

MB$SOFT$BAUD	EQU 0000$0100B	;software configurable baud rate

MB$SERIAL	EQU 0000$1000B	;serial enabled
MB$XON$XOFF	EQU 0001$0000B	;xon/xoff enabled

BAUD$NONE	EQU 0		;no baud rate
BAUD$50		EQU 1		;50 baud
BAUD$75		EQU 2		;75 baud
BAUD$110	EQU 3		;110 baud
BAUD$134	EQU 4		;134.5 baud
BAUD$150	EQU 5		;150 baud
BAUD$300	EQU 6		;300 baud
BAUD$600	EQU 7		;600 baud
BAUD$1200	EQU 8		;1200 baud
BAUD$1800	EQU 9		;1800 baud
BAUD$2400	EQU 10		;2400 baud
BAUD$3600	EQU 11		;3600 baud
BAUD$4800	EQU 12		;4800 baud
BAUD$7200	EQU 13		;7200 baud
BAUD$9600	EQU 14		;9600 baud
BAUD$19200	EQU 15		;19.2Kbaud

	CSEG

   IF	BANKED
?CINIT:
	LXI	H, BCINIT	;Banked CINIT
	JMP	BANKIO		;DISPATCHER

?CI:	LXI	H, BCI		;Banked CI
	JMP	BANKIO		;DISPATCHER

?CO:	LXI	H, BCO		;Banked CO
	JMP	BANKIO		;DISPATCHER

?CIST:	LXI	H, BCIST	;Banked CIST
	JMP	BANKIO		;DISPATCHER

?COST:	LXI	H, BCOST	;Banked COST

BANKIO:				;Banked DISPATCHER
;;;;;	SSPD	SPSAVE		;Z80 save stack pointer instruction
	LXI	H, 0		;8080 replacement: first, zero [HL]
	DAD	SP		;then add the [SP] to [HL] and then
	SHLD	SPSAVE		;save the current stack pointer
	LXI	SP, IOSP	;use local stack
	LDA	@CBNK		;get curent bank 
	PUSH	PSW		;save it on the stack
	XRA	A		;choose bank 0 (OP SYS)
	CALL	?BNKSL		;and select it
	LXI	D, BIORET	;place return address in [DE]
	PUSH	D		;and put it on the stack
	PCHL			;launch banked DISPATCHER
BIORET:
	POP	D		;get previous bank in [D]
	PUSH	PSW		;save handler return (if any)
	MOV	A, D		;put previous bank into [A]
	CALL	?BNKSL          ;and then reselect it
	POP	PSW		;put return code into [A]
;;;;;	LSPD	SPSAVE		;Z80 load stack pointer instruction
	LHLD	SPSAVE		;8080 replacement: Load into [HL]
	SPHL			;and then put [HL] into [SP]
	RET
   ENDIF

   IF BANKED
	DSEG			;Put banked handlers in banked segment
BCINIT:
   ELSE

?CINIT:				;?CINIT
   ENDIF
	MOV	B, C
	CALL	DEV$DISPATCH
	DW	CINIT0		;0
	DW	CINIT1		;1
	DW	CINIT2		;2
	DW	CINIT3		;3
	DW	NULL$INIT	;4
	DW	NULL$INIT	;5
	DW	NULL$INIT	;6
	DW	NULL$INIT	;7
	DW	NULL$INIT	;8
	DW	NULL$INIT	;9
	DW	NULL$INIT	;10
	DW	NULL$INIT	;11
	DW	NULL$INIT	;12
	DW	NULL$INIT	;13
	DW	NULL$INIT	;14
	DW	NULL$INIT	;15

   IF	BANKED
BCI:
   ELSE
?CI:				;CI
   ENDIF
	CALL	DEV$DISPATCH
	DW	CI0		;0
	DW	CI1		;1
	DW	CI2		;2
	DW	CI3		;3
	DW	NULL$CI		;4
	DW	NULL$CI		;5
	DW	NULL$CI		;6
	DW	NULL$CI		;7
	DW	NULL$CI		;8
	DW	NULL$CI		;9
	DW	NULL$CI		;10
	DW	NULL$CI		;11
	DW	NULL$CI		;12
	DW	NULL$CI		;13
	DW	NULL$CI		;14
	DW	NULL$CI		;15

   IF	BANKED
BCO:
   ELSE
?CO:				;?CO
   ENDIF
	CALL	DEV$DISPATCH
	DW	CO0		;0
	DW	CO1		;1
	DW	CO2		;2
	DW	CO3		;3
	DW	NULL$CO		;4
	DW	NULL$CO		;5
	DW	NULL$CO		;6
	DW	NULL$CO		;7
	DW	NULL$CO		;8
	DW	NULL$CO		;9
	DW	NULL$CO		;10
	DW	NULL$CO		;11
	DW	NULL$CO		;12
	DW	NULL$CO		;13
	DW	NULL$CO		;14
	DW	NULL$CO		;15

   IF	BANKED
BCIST:
   ELSE
?CIST:				;?CIST
   ENDIF
	mvi	a,UNDEF		;delayed-write FDC housecleaning
	sta	bfDrive
	call	ckFlush		;check FDC buffer while polling for user input
	ei
   IF	NOT MINIDSK
	mvi	a, cHDUNLD	;unload head
	out	DRVCMD
   ENDIF
	CALL	DEV$DISPATCH
	DW	CIST0		;0
	DW	CIST1		;1
	DW	CIST2		;2
	DW	CIST3		;3
	DW	NULL$CIST	;4
	DW	NULL$CIST	;5
	DW	NULL$CIST	;6
	DW	NULL$CIST	;7
	DW	NULL$CIST	;8
	DW	NULL$CIST	;9
	DW	NULL$CIST	;10
	DW	NULL$CIST	;11
	DW	NULL$CIST	;12
	DW	NULL$CIST	;13
	DW	NULL$CIST	;14
	DW	NULL$CIST	;15

   IF	BANKED
BCOST:
   ELSE
?COST:				;COST
   ENDIF
	CALL	DEV$DISPATCH
	DW	COST0		;0
	DW	COST1		;1
	DW	COST2		;2
	DW	COST3		;3
	DW	NULL$COST	;4
	DW	NULL$COST	;5
	DW	NULL$COST	;6
	DW	NULL$COST	;7
	DW	NULL$COST	;8
	DW	NULL$COST	;9
	DW	NULL$COST	;10
	DW	NULL$COST	;11
	DW	NULL$COST	;12
	DW	NULL$COST	;13
	DW	NULL$COST	;14
	DW	NULL$COST	;15

DEV$DISPATCH:			;Device handler
	MOV	A, B
	STA	DEV$CODE
	ADD	A		;x2 for word offset
	POP	H		;return address is 1st parameter
	MOV	E, A		;word offset in [DE]
	MVI	D, 0
	DAD	D		;add offset to calculate [HL] pointer
	MOV	E, M		;get handler address and put in [DE]
	INX	H
	MOV	D, M
	MOV	A, C		;put char in [A]
	XCHG			;put handler address into [HL]
	PCHL			;and launch it


;---------------------------------------------------------------------------
; Initialize serial ports 
;---------------------------------------------------------------------------

CINIT0:
	mvi	A, 03h		;reset 2SIO ports
	out	SIO1CTL
	out	SIO2CTL

	mvi	A, 11h		;select 8N2
	out	SIO1CTL
	out	SIO2CTL
CINIT1:
CINIT2:
CINIT3: 
	ret


;---------------------------------------------------------------------------
; Input status routines. Return FFh if character ready, else zero 
;---------------------------------------------------------------------------

CIST0:
sio1IS  in	SIO1CTL		;read 2SIO #1 status/control register
	ani	SIORDRF		;data present?
	rz			;no, return zero

	mvi	A, 0FFh		;else return FFh
	ret

CIST1:
sio2IS	in	SIO2CTL		;read 2SIO #2 status/control register
	ani	SIORDRF		;data present?
	rz			;no, return zero

	mvi	A, 0FFh		;else return FFh
	ret

CIST2:
sioIS	in	SIOCTL		;read SIO status/control register
	xri	0FFh		;convert to positive logic
	ani	SIORCV		;data present?
	rz			;no, return zero

	mvi	A, 0FFh		;else return FFh
	ret

CIST3:
acrIS	in	ACRCTL		;read ACR (SIO) control register
	xri	0FFh		;convert to positive logic
	ani	SIORCV		;data present?

	rz			;no, return zero
	mvi	A, 0FFh		;else return FFh
	ret


;---------------------------------------------------------------------------
; Character input routines (character ready assumed)
;---------------------------------------------------------------------------

CI0:
sio1In	in	SIO1DAT		;read and return the character
	ret

CI1:
sio2In	in	SIO2DAT		;read and return the character
	ret

CI2:
sioIn	in	SIODAT		;read and return the character
	ret

CI3:
acrIn	in	ACRDAT		;read and return the character
	ret


;---------------------------------------------------------------------------
; Output status routines. Return FFh if ready to send, else zero 
;---------------------------------------------------------------------------

COST0:
sio1OS	in	SIO1CTL		;read 2SIO #1 status/control register
	ani	SIOTDRE		;0=busy
	rz			;not ready, return 0

	mvi	A, 0FFh		;else return FFh
	ret

COST1:
sio2OS	in	SIO2CTL		;read 2SIO #2 status/control register
	ani	SIOTDRE		;0=busy
	rz			;not ready, return 0

	mvi	A, 0FFh		;else return FFh
	ret

COST2:
sioOS	in	SIOCTL		;read SIO status/control register
	xri	0FFh		;convert to positive logic
	ani	SIOXMT		;0=busy
	rz			;not ready, return 0

	mvi	A, 0FFh		;else return FFh
	ret

COST3:
acrOS	in	ACRCTL		;read SIO status/control register
	xri	0FFh		;convert to positive logic
	ani	SIOXMT		;0=busy
	rz			;not ready, return 0

	mvi	A, 0FFh		;else return FFh
	ret


;---------------------------------------------------------------------------
; Character output routines (transmit ready assumed)
;---------------------------------------------------------------------------

CO0:
sio1Out	out	SIO1DAT		;send character
	ret

CO1:
sio2Out	out	SIO2DAT		;send character
	ret

CO2:
sioOut	out	SIODAT		;send character
	ret

CO3:
acrOut	out	ACRDAT		;send character
	ret

NULL$CIST:			;null routines
NULL$COST:
	xra	A		;return \0
	jmp	NULL$RET
NULL$CI:
	mvi	A, 1Ah		;return ctrl-Z (EOF)
NULL$INIT:
NULL$CO:
NULL$RET:
	ret


;---------------------------------------------------------------------------
; Module Data
;---------------------------------------------------------------------------
DEV$CODE:
	DS	1

	CSEG
				;Character device table
@CTBL:
	DB '2SIOA '		;device 0, 2SIO at 10h/11h
	DB mb$in$out+mb$serial
	DB baud$none
	DB '2SIOB '		;device 1, 2SIO at 12h/13h
	DB mb$in$out+mb$serial
	DB baud$none
	DB 'SIO   '		;device 2, SIO at 0/1
	DB mb$in$out+mb$serial
	DB baud$none
	DB 'ACR   '		;device 3, SIO at 6/7 (ACR)
	DB mb$in$out+mb$serial
	DB baud$none
	DB 0

MAX$DEVICES	EQU	($-@CTBL)/8
	DB	0

	DS	24		;Character I/O local stack
IOSP	EQU	$
SPSAVE	DS	2

	END
